@@ -1,5 +1,5 @@
 /*
- * Copyright 2010-2014 The pygit2 contributors
+ * Copyright 2010-2024 The pygit2 contributors
  *
  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -27,9 +27,11 @@
 
 #define PY_SSIZE_T_CLEAN
 #include <Python.h>
+#include <git2.h>
 #include "error.h"
 #include "types.h"
 #include "utils.h"
+#include "odb.h"
 #include "oid.h"
 #include "repository.h"
 #include "object.h"
@@ -39,28 +41,53 @@ extern PyTypeObject CommitType;
 extern PyTypeObject BlobType;
 extern PyTypeObject TagType;
 
+PyTypeObject ObjectType;
 
 void
 Object_dealloc(Object* self)
 {
     Py_CLEAR(self->repo);
     git_object_free(self->obj);
+    git_tree_entry_free((git_tree_entry*)self->entry);
     Py_TYPE(self)->tp_free(self);
 }
 
 
+git_object*
+Object__load(Object *self)
+{
+    if (self->obj == NULL) {
+        int err = git_tree_entry_to_object(&self->obj, self->repo->repo, self->entry);
+        if (err < 0) {
+            Error_set(err);
+            return NULL;
+        }
+    }
+
+    return self->obj;
+}
+
+const git_oid*
+Object__id(Object *self)
+{
+    return (self->obj) ? git_object_id(self->obj) : git_tree_entry_id(self->entry);
+}
+
+
+git_object_t
+Object__type(Object *self)
+{
+    return (self->obj) ? git_object_type(self->obj) : git_tree_entry_type(self->entry);
+}
+
+
 PyDoc_STRVAR(Object_id__doc__,
     "The object id, an instance of the Oid type.");
 
 PyObject *
 Object_id__get__(Object *self)
 {
-    const git_oid *oid;
-
-    oid = git_object_id(self->obj);
-    assert(oid);
-
-    return git_oid_to_python(oid);
+    return git_oid_to_python(Object__id(self));
 }
 
 PyDoc_STRVAR(Object_oid__doc__,
@@ -75,29 +102,51 @@ Object_oid__get__(Object *self)
 
 PyDoc_STRVAR(Object_hex__doc__,
     "Hexadecimal representation of the object id. This is a shortcut for\n"
-    "Object.oid.hex\n"
+    "Object.id.hex\n"
     "This attribute is deprecated, please use 'id'\n");
 
 PyObject *
 Object_hex__get__(Object *self)
 {
-    const git_oid *oid;
+    return git_oid_to_py_str(Object__id(self));
+}
 
-    oid = git_object_id(self->obj);
-    assert(oid);
 
-    return git_oid_to_py_str(oid);
+PyDoc_STRVAR(Object_short_id__doc__,
+    "An unambiguous short (abbreviated) hex Oid string for the object.");
+
+PyObject *
+Object_short_id__get__(Object *self)
+{
+    if (Object__load(self) == NULL) { return NULL; } // Lazy load
+
+    git_buf short_id = { NULL, 0, 0 };
+    int err = git_object_short_id(&short_id, self->obj);
+    if (err != GIT_OK)
+        return Error_set(err);
+
+    PyObject *py_short_id = to_unicode_n(short_id.ptr, short_id.size, NULL, "strict");
+    git_buf_dispose(&short_id);
+    return py_short_id;
 }
 
 
 PyDoc_STRVAR(Object_type__doc__,
-    "One of the GIT_OBJ_COMMIT, GIT_OBJ_TREE, GIT_OBJ_BLOB or GIT_OBJ_TAG\n"
-    "constants.");
+    "One of the enums.ObjectType.COMMIT, TREE, BLOB or TAG constants.");
 
 PyObject *
 Object_type__get__(Object *self)
 {
-    return PyLong_FromLong(git_object_type(self->obj));
+    return PyLong_FromLong(Object__type(self));
+}
+
+PyDoc_STRVAR(Object_type_str__doc__,
+    "One of the 'commit', 'tree', 'blob' or 'tag' strings.");
+
+PyObject *
+Object_type_str__get__(Object *self)
+{
+    return to_path(git_object_type2string(Object__type(self)));
 }
 
 PyDoc_STRVAR(Object__pointer__doc__, "Get the object's pointer. For internal use only.");
@@ -105,25 +154,63 @@ PyObject *
 Object__pointer__get__(Object *self)
 {
     /* Bytes means a raw buffer */
+    if (Object__load(self) == NULL) { return NULL; } // Lazy load
     return PyBytes_FromStringAndSize((char *) &self->obj, sizeof(git_object *));
 }
 
+PyDoc_STRVAR(Object_name__doc__,
+             "Name (will be None if the object was not reached trough a tree)");
+PyObject *
+Object_name__get__(Object *self)
+{
+    if (self->entry == NULL)
+        Py_RETURN_NONE;
+
+    return to_path(git_tree_entry_name(self->entry));
+}
+
+PyDoc_STRVAR(Object_raw_name__doc__, "Name (bytes).");
+
+PyObject *
+Object_raw_name__get__(Object *self)
+{
+    if (self->entry == NULL)
+        Py_RETURN_NONE;
+
+    return PyBytes_FromString(git_tree_entry_name(self->entry));
+}
+
+PyDoc_STRVAR(Object_filemode__doc__,
+             "Filemode (will be None if the object was not reached trough a tree)");
+PyObject *
+Object_filemode__get__(Object *self)
+{
+    if (self->entry == NULL)
+        Py_RETURN_NONE;
+
+    return PyLong_FromLong(git_tree_entry_filemode(self->entry));
+}
+
 
 PyDoc_STRVAR(Object_read_raw__doc__,
-  "read_raw()\n"
+  "read_raw() -> bytes\n"
   "\n"
   "Returns the byte string with the raw contents of the object.");
 
 PyObject *
 Object_read_raw(Object *self)
 {
-    const git_oid *oid;
-    git_odb_object *obj;
+    int err;
+    git_odb *odb;
     PyObject *aux;
 
-    oid = git_object_id(self->obj);
+    err = git_repository_odb(&odb, self->repo->repo);
+    if (err < 0)
+        return Error_set(err);
 
-    obj = Repository_read_raw(self->repo->repo, oid, GIT_OID_HEXSZ);
+    const git_oid *oid = Object__id(self);
+    git_odb_object *obj = Odb_read_raw(odb, oid, GIT_OID_HEXSZ);
+    git_odb_free(odb);
     if (obj == NULL)
         return NULL;
 
@@ -135,61 +222,105 @@ Object_read_raw(Object *self)
     return aux;
 }
 
-static git_otype
-py_type_to_git_type(PyTypeObject *py_type)
-{
-    git_otype type = GIT_OBJ_BAD;
-
-    if (py_type == &CommitType) {
-        type = GIT_OBJ_COMMIT;
-    } else if (py_type == &TreeType) {
-        type = GIT_OBJ_TREE;
-    } else if (py_type == &BlobType) {
-        type = GIT_OBJ_BLOB;
-    } else if (py_type == &TagType) {
-        type = GIT_OBJ_TAG;
-    }
-
-    return type;
-}
-
 PyDoc_STRVAR(Object_peel__doc__,
   "peel(target_type) -> Object\n"
   "\n"
-  "Peel the current object and returns the first object of the given type\n");
+  "Peel the current object and returns the first object of the given type.\n"
+  "\n"
+  "If you pass None as the target type, then the object will be peeled\n"
+  "until the type changes. A tag will be peeled until the referenced object\n"
+  "is no longer a tag, and a commit will be peeled to a tree. Any other\n"
+  "object type will raise InvalidSpecError.\n");
 
 PyObject *
 Object_peel(Object *self, PyObject *py_type)
 {
-    int type = -1, err;
+    int err;
+    git_otype otype;
     git_object *peeled;
 
-    if (PyLong_Check(py_type)) {
-        type = PyLong_AsLong(py_type);
-        if (type == -1 && PyErr_Occurred())
-            return NULL;
-    } else if (PyType_Check(py_type)) {
-        type = py_type_to_git_type((PyTypeObject *) py_type);
-    }
+    if (Object__load(self) == NULL) { return NULL; } // Lazy load
 
-    if (type == -1) {
-        PyErr_SetString(PyExc_ValueError, "invalid target type");
+    otype = py_object_to_otype(py_type);
+    if (otype == GIT_OBJECT_INVALID)
         return NULL;
-    }
 
-    err = git_object_peel(&peeled, self->obj, (git_otype)type);
+    err = git_object_peel(&peeled, self->obj, otype);
     if (err < 0)
         return Error_set(err);
 
-    return wrap_object(peeled, self->repo);
+    return wrap_object(peeled, self->repo, NULL);
+}
+
+Py_hash_t
+Object_hash(Object *self)
+{
+    const git_oid *oid = Object__id(self);
+    PyObject *py_oid = git_oid_to_py_str(oid);
+    Py_hash_t ret = PyObject_Hash(py_oid);
+    Py_DECREF(py_oid);
+    return ret;
+}
+
+PyObject *
+Object_repr(Object *self)
+{
+    char hex[GIT_OID_HEXSZ + 1];
+
+    git_oid_fmt(hex, Object__id(self));
+    hex[GIT_OID_HEXSZ] = '\0';
+
+    return PyUnicode_FromFormat("<pygit2.Object{%s:%s}>",
+        git_object_type2string(Object__type(self)),
+        hex
+    );
+}
+
+PyObject *
+Object_richcompare(PyObject *o1, PyObject *o2, int op)
+{
+    PyObject *res;
+
+    if (!PyObject_TypeCheck(o2, &ObjectType)) {
+        Py_INCREF(Py_NotImplemented);
+        return Py_NotImplemented;
+    }
+
+    int equal = git_oid_equal(Object__id((Object *)o1), Object__id((Object *)o2));
+    switch (op) {
+        case Py_NE:
+            res = (equal) ? Py_False : Py_True;
+            break;
+        case Py_EQ:
+            res = (equal) ? Py_True : Py_False;
+            break;
+        case Py_LT:
+        case Py_LE:
+        case Py_GT:
+        case Py_GE:
+            Py_INCREF(Py_NotImplemented);
+            return Py_NotImplemented;
+        default:
+            PyErr_Format(PyExc_RuntimeError, "Unexpected '%d' op", op);
+            return NULL;
+    }
+
+    Py_INCREF(res);
+    return res;
 }
 
 PyGetSetDef Object_getseters[] = {
     GETTER(Object, oid),
     GETTER(Object, id),
     GETTER(Object, hex),
+    GETTER(Object, short_id),
     GETTER(Object, type),
+    GETTER(Object, type_str),
     GETTER(Object, _pointer),
+    // These come from git_tree_entry
+    GETTER(Object, name),
+    GETTER(Object, raw_name),
+    GETTER(Object, filemode),
     {NULL}
 };
 
@@ -212,11 +343,11 @@ PyTypeObject ObjectType = {
     0,                                         /* tp_getattr        */
     0,                                         /* tp_setattr        */
     0,                                         /* tp_compare        */
-    0,                                         /* tp_repr           */
+    (reprfunc)Object_repr,                     /* tp_repr           */
     0,                                         /* tp_as_number      */
     0,                                         /* tp_as_sequence    */
     0,                                         /* tp_as_mapping     */
-    0,                                         /* tp_hash           */
+    (hashfunc)Object_hash,                     /* tp_hash           */
     0,                                         /* tp_call           */
     0,                                         /* tp_str            */
     0,                                         /* tp_getattro       */
@@ -226,7 +357,7 @@ PyTypeObject ObjectType = {
     Object__doc__,                             /* tp_doc            */
     0,                                         /* tp_traverse       */
     0,                                         /* tp_clear          */
-    0,                                         /* tp_richcompare    */
+    (richcmpfunc)Object_richcompare,           /* tp_richcompare    */
     0,                                         /* tp_weaklistoffset */
     0,                                         /* tp_iter           */
     0,                                         /* tp_iternext       */
@@ -244,21 +375,23 @@ PyTypeObject ObjectType = {
 };
 
 PyObject *
-wrap_object(git_object *c_object, Repository *repo)
+wrap_object(git_object *c_object, Repository *repo, const git_tree_entry *entry)
 {
     Object *py_obj = NULL;
 
-    switch (git_object_type(c_object)) {
-        case GIT_OBJ_COMMIT:
+    git_object_t obj_type = (c_object) ? git_object_type(c_object) : git_tree_entry_type(entry);
+
+    switch (obj_type) {
+        case GIT_OBJECT_COMMIT:
             py_obj = PyObject_New(Object, &CommitType);
             break;
-        case GIT_OBJ_TREE:
+        case GIT_OBJECT_TREE:
             py_obj = PyObject_New(Object, &TreeType);
             break;
-        case GIT_OBJ_BLOB:
+        case GIT_OBJECT_BLOB:
             py_obj = PyObject_New(Object, &BlobType);
             break;
-        case GIT_OBJ_TAG:
+        case GIT_OBJECT_TAG:
             py_obj = PyObject_New(Object, &TagType);
             break;
         default:
@@ -271,6 +404,7 @@ wrap_object(git_object *c_object, Repository *repo)
             py_obj->repo = repo;
             Py_INCREF(repo);
         }
+        py_obj->entry = entry;
     }
     return (PyObject *)py_obj;
 }
