@@ -1,5 +1,5 @@
 /*
- * Copyright 2010-2014 The pygit2 contributors
+ * Copyright 2010-2024 The pygit2 contributors
  *
  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -36,12 +36,15 @@
 #include "oid.h"
 #include "signature.h"
 #include "reference.h"
+#include <git2/sys/refs.h>
 
 
 extern PyObject *GitError;
 extern PyTypeObject RefLogEntryType;
+extern PyTypeObject RepositoryType;
 extern PyTypeObject SignatureType;
 
+PyTypeObject ReferenceType;
 
 void RefLogIter_dealloc(RefLogIter *self)
 {
@@ -53,16 +56,20 @@ PyObject *
 RefLogIter_iternext(RefLogIter *self)
 {
     const git_reflog_entry *entry;
+    const char * entry_message;
     RefLogEntry *py_entry;
     int err;
 
     if (self->i < self->size) {
         entry = git_reflog_entry_byindex(self->reflog, self->i);
         py_entry = PyObject_New(RefLogEntry, &RefLogEntryType);
+        if (py_entry == NULL)
+            return NULL;
 
-        py_entry->oid_old = git_oid_allocfmt(git_reflog_entry_id_old(entry));
-        py_entry->oid_new = git_oid_allocfmt(git_reflog_entry_id_new(entry));
-        py_entry->message = strdup(git_reflog_entry_message(entry));
+        py_entry->oid_old = git_oid_to_python(git_reflog_entry_id_old(entry));
+        py_entry->oid_new = git_oid_to_python(git_reflog_entry_id_new(entry));
+        entry_message = git_reflog_entry_message(entry);
+        py_entry->message = (entry_message != NULL) ? strdup(entry_message) : NULL;
         err = git_signature_dup(&py_entry->signature,
                                 git_reflog_entry_committer(entry));
         if (err < 0)
@@ -110,6 +117,62 @@ PyTypeObject RefLogIterType = {
     (iternextfunc)RefLogIter_iternext,         /* tp_iternext       */
 };
 
+PyDoc_STRVAR(Reference__doc__,
+        "Reference(name: str, target: str): create a symbolic reference\n"
+        "\n"
+        "Reference(name: str, oid: Oid, peel: Oid): create a direct reference\n"
+        "\n"
+        "'peel' is the first non-tag object's OID, or None.\n"
+        "\n"
+        "The purpose of this constructor is for use in custom refdb backends.\n"
+        "References created with this function are unlikely to work as\n"
+        "expected in other contexts.\n");
+
+static int
+Reference_init_symbolic(Reference *self, PyObject *args, PyObject *kwds)
+{
+    const char *name, *target;
+    if (!PyArg_ParseTuple(args, "ss", &name, &target)) {
+        return -1;
+    }
+    self->reference = git_reference__alloc_symbolic(name, target);
+    return 0;
+}
+
+int
+Reference_init(Reference *self, PyObject *args, PyObject *kwds)
+{
+    const char *name;
+    git_oid oid, peel;
+    PyObject *py_oid, *py_peel;
+
+    if (kwds && PyDict_Size(kwds) > 0) {
+        PyErr_SetString(PyExc_TypeError, "Reference takes no keyword arguments");
+        return -1;
+    }
+
+    Py_ssize_t nargs = PyTuple_Size(args);
+    if (nargs == 2) {
+        return Reference_init_symbolic(self, args, kwds);
+    } else if (nargs != 3) {
+        PyErr_SetString(PyExc_TypeError,
+                "Invalid arguments to Reference constructor");
+        return -1;
+    }
+
+    if (!PyArg_ParseTuple(args, "sOO", &name, &py_oid, &py_peel)) {
+        return -1;
+    }
+
+    py_oid_to_git_oid(py_oid, &oid);
+    if (py_peel != Py_None) {
+        py_oid_to_git_oid(py_peel, &peel);
+    }
+
+    self->reference = git_reference__alloc(name, &oid, &peel);
+    return 0;
+}
+
 void
 Reference_dealloc(Reference *self)
 {
@@ -144,32 +207,31 @@ Reference_delete(Reference *self, PyObject *args)
 
 
 PyDoc_STRVAR(Reference_rename__doc__,
-  "rename(new_name)\n"
+  "rename(new_name: str)\n"
   "\n"
   "Rename the reference.");
 
 PyObject *
 Reference_rename(Reference *self, PyObject *py_name)
 {
-    char *c_name;
-    int err;
-    git_reference *new_reference;
-
     CHECK_REFERENCE(self);
 
-    /* Get the C name */
-    c_name = py_path_to_c_str(py_name);
+    // Get the C name
+    char *c_name = pgit_encode_fsdefault(py_name);
     if (c_name == NULL)
         return NULL;
 
-    /* Rename */
-    err = git_reference_rename(&new_reference, self->reference, c_name, 0, NULL, NULL);
-    git_reference_free(self->reference);
+    // Rename
+    git_reference *new_reference;
+    int err = git_reference_rename(&new_reference, self->reference, c_name, 0, NULL);
     free(c_name);
-    if (err < 0)
+    if (err)
         return Error_set(err);
 
+    // Upadate reference
+    git_reference_free(self->reference);
     self->reference = new_reference;
+
     Py_RETURN_NONE;
 }
 
@@ -202,6 +264,23 @@ Reference_resolve(Reference *self, PyObject *args)
 }
 
 
+static PyObject *
+Reference_target_impl(Reference *self, const char ** c_name)
+{
+    CHECK_REFERENCE(self);
+
+    /* Case 1: Direct */
+    if (GIT_REF_OID == git_reference_type(self->reference))
+        return git_oid_to_python(git_reference_target(self->reference));
+
+    /* Case 2: Symbolic */
+    *c_name = git_reference_symbolic_target(self->reference);
+    if (*c_name == NULL)
+        PyErr_SetString(PyExc_ValueError, "no target available");
+
+    return NULL;
+}
+
 PyDoc_STRVAR(Reference_target__doc__,
     "The reference target: If direct the value will be an Oid object, if it\n"
     "is symbolic it will be an string with the full name of the target\n"
@@ -210,39 +289,47 @@ PyDoc_STRVAR(Reference_target__doc__,
 PyObject *
 Reference_target__get__(Reference *self)
 {
-    const char * c_name;
+    const char * c_name = NULL;
+    PyObject * ret;
 
-    CHECK_REFERENCE(self);
+    ret = Reference_target_impl(self, &c_name);
+    if (ret != NULL)
+        return ret;
+    if (c_name != NULL)
+        return to_path(c_name);
+    return NULL;
+}
 
-    /* Case 1: Direct */
-    if (GIT_REF_OID == git_reference_type(self->reference))
-        return git_oid_to_python(git_reference_target(self->reference));
 
-    /* Case 2: Symbolic */
-    c_name = git_reference_symbolic_target(self->reference);
-    if (c_name == NULL) {
-        PyErr_SetString(PyExc_ValueError, "no target available");
-        return NULL;
-    }
-    return to_path(c_name);
+PyDoc_STRVAR(Reference_raw_target__doc__,
+    "The raw reference target: If direct the value will be an Oid object, if it\n"
+    "is symbolic it will be bytes with the full name of the target\n"
+    "reference.\n");
+
+PyObject *
+Reference_raw_target__get__(Reference *self)
+{
+    const char * c_name = NULL;
+    PyObject * ret;
+
+    ret = Reference_target_impl(self, &c_name);
+    if (ret != NULL)
+        return ret;
+    if (c_name != NULL)
+        return PyBytes_FromString(c_name);
+    return NULL;
 }
 
 PyDoc_STRVAR(Reference_set_target__doc__,
-    "set_target(target, [signature, message])\n"
+    "set_target(target[, message: str])\n"
     "\n"
-    "Set the target of this reference.\n"
+    "Set the target of this reference. Creates a new entry in the reflog.\n"
     "\n"
-    "Update the reference using the given signature and message.\n"
-    "These will be used to fill the reflog entry which will be created\n"
-    "as a result of this update\n"
-    "\n"
-    "Arguments:\n"
+    "Parameters:\n"
     "\n"
     "target\n"
     "    The new target for this reference\n"
-    "signature\n"
-    "    The signature to use for the reflog. If left out, the repository's\n"
-    "    default identity will be used.\n"
+    "\n"
     "message\n"
     "    Message to use for the reflog.\n");
 
@@ -250,31 +337,25 @@ PyObject *
 Reference_set_target(Reference *self, PyObject *args, PyObject *kwds)
 {
     git_oid oid;
-    char *c_name;
     int err;
     git_reference *new_ref;
-    const git_signature *sig = NULL;
     PyObject *py_target = NULL;
-    Signature *py_signature = NULL;
     const char *message = NULL;
-    char *keywords[] = {"target", "signature", "message", NULL};
+    char *keywords[] = {"target", "message", NULL};
 
     CHECK_REFERENCE(self);
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O!s", keywords,
-                                     &py_target, &SignatureType, &py_signature, &message))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|s", keywords,
+                                     &py_target, &message))
         return NULL;
 
-    if (py_signature)
-        sig = py_signature->signature;
-
     /* Case 1: Direct */
     if (GIT_REF_OID == git_reference_type(self->reference)) {
         err = py_oid_to_git_oid_expand(self->repo->repo, py_target, &oid);
         if (err < 0)
             goto error;
 
-        err = git_reference_set_target(&new_ref, self->reference, &oid, sig, message);
+        err = git_reference_set_target(&new_ref, self->reference, &oid, message);
         if (err < 0)
             goto error;
 
@@ -284,11 +365,11 @@ Reference_set_target(Reference *self, PyObject *args, PyObject *kwds)
     }
 
     /* Case 2: Symbolic */
-    c_name = py_path_to_c_str(py_target);
+    char *c_name = pgit_encode_fsdefault(py_target);
     if (c_name == NULL)
         return NULL;
 
-    err = git_reference_symbolic_set_target(&new_ref, self->reference, c_name, sig, message);
+    err = git_reference_symbolic_set_target(&new_ref, self->reference, c_name, message);
     free(c_name);
     if (err < 0)
         goto error;
@@ -313,6 +394,15 @@ Reference_name__get__(Reference *self)
     return to_path(git_reference_name(self->reference));
 }
 
+PyDoc_STRVAR(Reference_raw_name__doc__, "The full name of the reference (Bytes).");
+
+PyObject *
+Reference_raw_name__get__(Reference *self)
+{
+    CHECK_REFERENCE(self);
+    return PyBytes_FromString(git_reference_name(self->reference));
+}
+
 PyDoc_STRVAR(Reference_shorthand__doc__,
     "The shorthand \"human-readable\" name of the reference.");
 
@@ -323,13 +413,23 @@ Reference_shorthand__get__(Reference *self)
     return to_path(git_reference_shorthand(self->reference));
 }
 
+PyDoc_STRVAR(Reference_raw_shorthand__doc__,
+    "The shorthand \"human-readable\" name of the reference (Bytes).");
+
+PyObject *
+Reference_raw_shorthand__get__(Reference *self)
+{
+    CHECK_REFERENCE(self);
+    return PyBytes_FromString(git_reference_shorthand(self->reference));
+}
+
 PyDoc_STRVAR(Reference_type__doc__,
     "Type, either GIT_REF_OID or GIT_REF_SYMBOLIC.");
 
 PyObject *
 Reference_type__get__(Reference *self)
 {
-    git_ref_t c_type;
+    git_reference_t c_type;
 
     CHECK_REFERENCE(self);
     c_type = git_reference_type(self->reference);
@@ -364,26 +464,100 @@ Reference_log(Reference *self)
     return (PyObject*)iter;
 }
 
-PyDoc_STRVAR(Reference_get_object__doc__,
-  "get_object() -> object\n"
+PyDoc_STRVAR(Reference_peel__doc__,
+  "peel(type=None) -> Object\n"
   "\n"
-  "Retrieves the object the current reference is pointing to.");
+  "Retrieve an object of the given type by recursive peeling.\n"
+  "\n"
+  "If no type is provided, the first non-tag object will be returned.");
 
 PyObject *
-Reference_get_object(Reference *self)
+Reference_peel(Reference *self, PyObject *args)
 {
     int err;
-    git_object* obj;
+    git_otype otype;
+    git_object *obj;
+    PyObject *py_type = Py_None;
 
     CHECK_REFERENCE(self);
 
-    err = git_reference_peel(&obj, self->reference, GIT_OBJ_ANY);
+    if (!PyArg_ParseTuple(args, "|O", &py_type))
+        return NULL;
+
+    otype = py_object_to_otype(py_type);
+    if (otype == GIT_OBJECT_INVALID)
+        return NULL;
+
+    err = git_reference_peel(&obj, self->reference, otype);
     if (err < 0)
         return Error_set(err);
 
-    return wrap_object(obj, self->repo);
+    return wrap_object(obj, self->repo, NULL);
 }
 
+PyObject *
+Reference_richcompare(PyObject *o1, PyObject *o2, int op)
+{
+    PyObject *res;
+    Reference *obj1;
+    Reference *obj2;
+    const char *name1;
+    const char *name2;
+
+    if (!PyObject_TypeCheck(o2, &ReferenceType)) {
+        Py_INCREF(Py_NotImplemented);
+        return Py_NotImplemented;
+    }
+
+    switch (op) {
+        case Py_NE:
+            obj1 = (Reference *) o1;
+            obj2 = (Reference *) o2;
+
+            CHECK_REFERENCE(obj1);
+            CHECK_REFERENCE(obj2);
+
+            name1 = git_reference_name(obj1->reference);
+            name2 = git_reference_name(obj2->reference);
+
+            if (strcmp(name1, name2) != 0) {
+                res = Py_True;
+                break;
+            }
+
+            res = Py_False;
+            break;
+        case Py_EQ:
+            obj1 = (Reference *) o1;
+            obj2 = (Reference *) o2;
+
+            CHECK_REFERENCE(obj1);
+            CHECK_REFERENCE(obj2);
+
+            name1 = git_reference_name(obj1->reference);
+            name2 = git_reference_name(obj2->reference);
+
+            if (strcmp(name1, name2) != 0) {
+                res = Py_False;
+                break;
+            }
+
+            res = Py_True;
+            break;
+        case Py_LT:
+        case Py_LE:
+        case Py_GT:
+        case Py_GE:
+            Py_INCREF(Py_NotImplemented);
+            return Py_NotImplemented;
+        default:
+            PyErr_Format(PyExc_RuntimeError, "Unexpected '%d' op", op);
+            return NULL;
+    }
+
+    Py_INCREF(res);
+    return res;
+}
 
 PyDoc_STRVAR(RefLogEntry_committer__doc__, "Committer.");
 
@@ -409,16 +583,16 @@ RefLogEntry_init(RefLogEntry *self, PyObject *args, PyObject *kwds)
 static void
 RefLogEntry_dealloc(RefLogEntry *self)
 {
-    free(self->oid_old);
-    free(self->oid_new);
+    Py_CLEAR(self->oid_old);
+    Py_CLEAR(self->oid_new);
     free(self->message);
     git_signature_free(self->signature);
     PyObject_Del(self);
 }
 
 PyMemberDef RefLogEntry_members[] = {
-    MEMBER(RefLogEntry, oid_new, T_STRING, "New oid."),
-    MEMBER(RefLogEntry, oid_old, T_STRING, "Old oid."),
+    MEMBER(RefLogEntry, oid_new, T_OBJECT, "New oid."),
+    MEMBER(RefLogEntry, oid_old, T_OBJECT, "Old oid."),
     MEMBER(RefLogEntry, message, T_STRING, "Message."),
     {NULL}
 };
@@ -477,22 +651,22 @@ PyMethodDef Reference_methods[] = {
     METHOD(Reference, rename, METH_O),
     METHOD(Reference, resolve, METH_NOARGS),
     METHOD(Reference, log, METH_NOARGS),
-    METHOD(Reference, get_object, METH_NOARGS),
     METHOD(Reference, set_target, METH_VARARGS | METH_KEYWORDS),
+    METHOD(Reference, peel, METH_VARARGS),
     {NULL}
 };
 
 PyGetSetDef Reference_getseters[] = {
     GETTER(Reference, name),
+    GETTER(Reference, raw_name),
     GETTER(Reference, shorthand),
+    GETTER(Reference, raw_shorthand),
     GETTER(Reference, target),
+    GETTER(Reference, raw_target),
     GETTER(Reference, type),
     {NULL}
 };
 
-
-PyDoc_STRVAR(Reference__doc__, "Reference.");
-
 PyTypeObject ReferenceType = {
     PyVarObject_HEAD_INIT(NULL, 0)
     "_pygit2.Reference",                       /* tp_name           */
@@ -517,7 +691,7 @@ PyTypeObject ReferenceType = {
     Reference__doc__,                          /* tp_doc            */
     0,                                         /* tp_traverse       */
     0,                                         /* tp_clear          */
-    0,                                         /* tp_richcompare    */
+    Reference_richcompare,                     /* tp_richcompare    */
     0,                                         /* tp_weaklistoffset */
     0,                                         /* tp_iter           */
     0,                                         /* tp_iternext       */
@@ -529,7 +703,7 @@ PyTypeObject ReferenceType = {
     0,                                         /* tp_descr_get      */
     0,                                         /* tp_descr_set      */
     0,                                         /* tp_dictoffset     */
-    0,                                         /* tp_init           */
+    (initproc)Reference_init,                  /* tp_init           */
     0,                                         /* tp_alloc          */
     0,                                         /* tp_new            */
 };
@@ -543,8 +717,8 @@ wrap_reference(git_reference * c_reference, Repository *repo)
     py_reference = PyObject_New(Reference, &ReferenceType);
     if (py_reference) {
         py_reference->reference = c_reference;
+        py_reference->repo = repo;
         if (repo) {
-            py_reference->repo = repo;
             Py_INCREF(repo);
         }
     }
